@@ -8,7 +8,7 @@ import defopt
 import os
 from glob import glob
 from typing import List, Optional
-from . import data, models, utils, predict, io, evaluate  #, timeseries
+from . import data, models, utils, predict, io, evaluate, neptune, data_hash  #, timeseries
 
 try:
     from tensorflow.python.framework.ops import disable_eager_execution
@@ -22,14 +22,15 @@ def train(*, data_dir: str, y_suffix: str = '',
           model_name: str = 'tcn', nb_filters: int = 16, kernel_size: int = 16,
           nb_conv: int = 3, use_separable: List[bool] = False, nb_hist: int = 1024,
           ignore_boundaries: bool = True, batch_norm: bool = True,
-          nb_pre_conv: int = 0,
+          nb_pre_conv: int = 0, pre_nb_dft: int = 64,
           pre_kernel_size: int = 3, pre_nb_filters: int = 16, pre_nb_conv: int = 2,
           verbose: int = 2, batch_size: int = 32,
           nb_epoch: int = 400,
           learning_rate: Optional[float] = None, reduce_lr: bool = False, reduce_lr_patience: int = 5,
           fraction_data: Optional[float] = None, seed: Optional[int] = None, batch_level_subsampling: bool = False,
-          tensorboard: bool = False, log_messages: bool = False,
-          nb_stacks: int = 2, with_y_hist: bool = True, x_suffix: str = '',
+          tensorboard: bool = False, neptune_api_token: Optional[str] = None, neptune_project: Optional[str] = None,
+          log_messages: bool = False, nb_stacks: int = 2, with_y_hist: bool = True, x_suffix: str = '',
+          balance: bool = False, version_data: bool = False,
           _qt_progress: bool = False):
     """Train a DeepSS network.
 
@@ -68,13 +69,16 @@ def train(*, data_dir: str, y_suffix: str = '',
                                   Defaults to True.
         batch_norm (bool): Batch normalize.
                            Defaults to True.
-        nb_pre_conv (int): Adds downsampling frontend.
-                           TCN: adds a frontend of N conv blocks (conv-relu-batchnorm-maxpool2) to the TCN - useful for reducing the sampling rate for USV.
-                           TCN_STFT: stft
+        nb_pre_conv (int): Downsampling rate. Adds downsampling frontend if not 0.
+                           TCN_TCN: adds a frontend of N conv blocks (conv-relu-batchnorm-maxpool2) to the TCN.
+                           TCN_STFT: adds a trainable STFT frontend.
                            Defaults to 0 (no frontend).
-        pre_nb_filters (int): [description]. Defaults to 16.
-        pre_kernel_size (int): [description]. Defaults to 3.
-        pre_nb_conv (int): [description]. Defaults to 3.
+        pre_nb_dft (int): Number of filters (roughly corresponding to filters) in the STFT frontend.
+                          Defaults to 64.
+        pre_nb_filters (int): Number of filters per layer in the pre-processing TCN.
+                              Defaults to 16.
+        pre_kernel_size (int): Duration of filters (=kernels) in samples in the pre-processing TCN.
+                               Defaults to 3.
         verbose (int): Verbosity of training output (0 - no output(?), 1 - progress bar, 2 - one line per epoch).
                        Defaults to 2.
         batch_size (int): Batch size
@@ -97,16 +101,28 @@ def train(*, data_dir: str, y_suffix: str = '',
         batch_level_subsampling (bool): Select fraction of data for training from random subset of shuffled batches.
                                         If False, select a continuous chunk of the recording.
                                         Defaults to False.
-        tensorboard (bool): Write tensorboard logs to save_dir.
-                            Defaults to False.
+        tensorboard (bool): Write tensorboard logs to save_dir. Defaults to False.
+
+        neptune_api_token (Optional[str]): API token for logging to neptune.ai. Defaults to None (no logging).
+        neptune_project (Optional[str]): Project to log to for neptune.ai. Defaults to None (no logging).
         log_messages (bool): Sets logging level to INFO.
                              Defaults to False (will follow existing settings).
         nb_stacks (int): Unused if model name is "tcn" or "tcn_stft". Defaults to 2.
         with_y_hist (bool): Unused if model name is "tcn" or "tcn_stft". Defaults to True.
         x_suffix (str): Select specific training data based on suffix (e.g. x_suffix).
                         Defaults to '' (will use the standard data 'x')
-        """
+        balance (bool): Balance data. Weights class-wise errors by the inverse of the class frequencies.
+                        Defaults to False.
+        version_data (bool): Save MD5 hash of the data_dir to log and params.yaml.
+                             Defaults to False (can take long for large datasets ).
 
+        Returns
+            model (tf.keras.Model)
+            params (Dict[str:Any])
+        """
+        # _qt_progress: tuple of (multiprocessing.Queue, threading.Event)
+        #        The queue is used to transmit progress updates to the GUI,
+        #        the event is set in the GUI to stop training.
     if log_messages:
         logging.basicConfig(level=logging.INFO)
 
@@ -146,16 +162,23 @@ def train(*, data_dir: str, y_suffix: str = '',
     if '_multi' in model_name:
         params['unpack_channels'] = True
 
-    logging.info('loading data')
+    logging.info('Loading data from {data_dir}.')
     d = io.load(data_dir, x_suffix=x_suffix, y_suffix=y_suffix)
+
     params.update(d.attrs)  # add metadata from data.attrs to params for saving
 
+    if version_data:
+        params['data_hash'] = data_hash.hash_data(data_dir)
+        logging.info(f"Version of the data:")
+        logging.info(f"   MD5 hash of {data_dir} is")
+        logging.info(f"   {params['data_hash']}")
+
     if fraction_data is not None:
         if fraction_data > 1.0:  # seconds
             logging.info(f"{fraction_data} seconds corresponds to {fraction_data / (d['train']['x'].shape[0] / d.attrs['samplerate_x_Hz']):1.4f} of the training data.")
             fraction_data = np.min((fraction_data / (d['train']['x'].shape[0] / d.attrs['samplerate_x_Hz']), 1.0))
         elif fraction_data < 1.0:
-            logging.info(f"Using {fraction_data:1.4f} of data for training and validation.")
+            logging.info(f"Using {fraction_data:1.4f} of the training and validation data.")
 
     if fraction_data is not None and not batch_level_subsampling:  # train on a subset
         min_nb_samples = nb_hist * (batch_size + 2)  # ensure the generator contains at least one full batch
@@ -173,7 +196,7 @@ def train(*, data_dir: str, y_suffix: str = '',
     logging.info('Parameters:')
     logging.info(params)
 
-    logging.info('preparing data')
+    logging.info('Preparing data')
     if fraction_data is not None and batch_level_subsampling:  # train on a subset
         np.random.seed(seed)
         shuffle_subset = fraction_data
@@ -197,10 +220,19 @@ def train(*, data_dir: str, y_suffix: str = '',
     #                              shuffle=False, batch_size=batch_size,
     #                              start_index=first_sample_val, end_index=last_sample_val)
 
-    logging.info('Training data:')
-    logging.info(data_gen)
-    logging.info('Validation data:')
-    logging.info(val_gen)
+    logging.info(f"Training data:")
+    logging.info(f"   {data_gen}")
+    logging.info(f"Validation data:")
+    logging.info(f"   {val_gen}")
+
+    params['class_weights'] = None
+    if balance:
+        from sklearn.utils import class_weight
+        y_train = np.argmax(d['train']['y'], axis=1)
+        params['class_weights'] = class_weight.compute_class_weight('balanced',
+                                                        np.unique(y_train),
+                                                        y_train)
+        logging.info(f'Balancing classes: {class_weights}')
 
     logging.info('building network')
     model = models.model_dict[model_name](**params)
@@ -219,12 +251,23 @@ def train(*, data_dir: str, y_suffix: str = '',
     if reduce_lr:
         callbacks.append(ReduceLROnPlateau(patience=reduce_lr_patience, verbose=1))
 
-    if _qt_progress is not None:
+    if _qt_progress:
         callbacks.append(utils.QtProgressCallback(nb_epoch, _qt_progress))
 
     if tensorboard:
         callbacks.append(TensorBoard(log_dir=save_dir))
 
+    del params['neptune_api_token']
+    if neptune_api_token and neptune_project:  # could also get those from env vars!
+        if not neptune.HAS_NEPTUNE:
+            logging.error('Could not import neptune in das.neptune.')
+        else:
+            try:
+                poseidon = neptune.Poseidon(neptune_project, neptune_api_token, params)
+                callbacks.append(poseidon.callback())
+            except Exception as e:
+                logging.exception('Neptune stuff failed.')
+
     # TRAIN NETWORK
     logging.info('start training')
     fit_hist = model.fit(
@@ -234,6 +277,7 @@ def train(*, data_dir: str, y_suffix: str = '',
         verbose=verbose,
         validation_data=val_gen,
         callbacks=callbacks,
+        class_weight=params['class_weights'],
     )
 
     # TEST
@@ -252,9 +296,11 @@ def train(*, data_dir: str, y_suffix: str = '',
         labels_pred = predict.labels_from_probabilities(y_pred)
 
         logging.info('evaluating')
-        conf_mat, report = evaluate.evaluate_segments(labels_test, labels_pred, params['class_names'])
+        conf_mat, report = evaluate.evaluate_segments(labels_test, labels_pred, params['class_names'], report_as_dict=True)
         logging.info(conf_mat)
         logging.info(report)
+        if neptune_api_token and neptune_project:  # could also get those from env vars!
+            poseidon.log_test_results(report)
 
         save_filename = "{0}_results.h5".format(save_name)
         logging.info('saving to ' + save_filename + '.')
@@ -270,3 +316,6 @@ def train(*, data_dir: str, y_suffix: str = '',
             }
 
         fl.save(save_filename, d)
+
+    logging.info('DONE.')
+    return model, params
